import ast
import optparse
from ast import NodeVisitor, Store
from typing import Iterable, List, Tuple, Union

import flake8.options.manager


FunctionTypes = Union[ast.AsyncFunctionDef, ast.FunctionDef, ast.Lambda]
LintResult = Tuple[int, int, str, str]


class Plugin:
    name = "flake8-unused-arguments"
    version = "0.0.2"

    ignore_abstract = False
    ignore_stubs = False
    ignore_variadic_names = False

    def __init__(self, tree: ast.Module):
        self.tree = tree

    @classmethod
    def add_options(cls, option_manager: flake8.options.manager.OptionManager) -> None:
        option_manager.add_option(
            "--unused-arguments-ignore-abstract-functions",
            action="store_true",
            parse_from_config=True,
            default=cls.ignore_abstract,
            dest="unused_arguments_ignore_abstract_functions",
            help="If provided, then unused arguments for functions decorated with abstractmethod will be ignored.",
        )

        option_manager.add_option(
            "--unused-arguments-ignore-stub-functions",
            action="store_true",
            parse_from_config=True,
            default=cls.ignore_stubs,
            dest="unused_arguments_ignore_stub_functions",
            help="If provided, then unused arguments for functions that are only a pass statement will be ignored.",
        )

        option_manager.add_option(
            "--unused-arguments-ignore-variadic-names",
            action="store_true",
            parse_from_config=True,
            default=cls.ignore_variadic_names,
            dest="unused_arguments_ignore_variadic_names",
            help="If provided, then unused *args and **kwargs won't produce warnings.",
        )

    @classmethod
    def parse_options(cls, options: optparse.Values) -> None:
        cls.ignore_abstract = options.unused_arguments_ignore_abstract_functions
        cls.ignore_stubs = options.unused_arguments_ignore_stub_functions
        cls.ignore_variadic_names = options.unused_arguments_ignore_variadic_names

    def run(self) -> Iterable[LintResult]:
        finder = FunctionFinder()
        finder.visit(self.tree)

        for function in finder.functions:
            decorator_names = set(get_decorator_names(function))
            # ignore abtractmethods, it's not a surprise when they're empty
            if self.ignore_abstract and "abstractmethod" in decorator_names:
                continue

            # ignore stub functions
            if self.ignore_stubs and is_stub_function(function):
                continue

            for i, name in enumerate(get_unused_arguments(function)):
                if self.ignore_variadic_names:
                    if function.args.vararg and function.args.vararg.arg == name:
                        continue
                    if function.args.kwarg and function.args.kwarg.arg == name:
                        continue

                # ignore self or whatever the first argument is for a classmethod
                if i == 0 and (name == "self" or "classmethod" in decorator_names):
                    continue

                line_number = function.lineno
                offset = function.col_offset

                if name.startswith("_"):
                    error_code = "U101"
                else:
                    error_code = "U100"

                text = "{error_code} Unused argument '{name}'".format(
                    error_code=error_code, name=name
                )
                check = "unused argument"
                yield (line_number, offset, text, check)


def get_unused_arguments(function: FunctionTypes) -> List[str]:
    """Generator that yields all of the unused arguments in the given function."""
    names = get_argument_names(function)

    class NameFinder(NodeVisitor):
        def visit_Name(self, name: ast.Name) -> None:
            if isinstance(name.ctx, Store):
                return

            if name.id in names:
                names.remove(name.id)

    NameFinder().visit(function)

    return names


def get_argument_names(function: FunctionTypes) -> List[str]:
    """Get all of the argument names of the given function."""
    args = function.args

    names: List[str] = []

    # plain old args
    names.extend(arg.arg for arg in args.args)

    # *arg name
    if args.vararg is not None:
        names.append(args.vararg.arg)

    # *, key, word, only, args
    names.extend(arg.arg for arg in args.kwonlyargs)

    # **kwarg name
    if args.kwarg is not None:
        names.append(args.kwarg.arg)

    return names


def get_decorator_names(function: FunctionTypes) -> Iterable[str]:
    if isinstance(function, ast.Lambda):
        return

    for decorator in function.decorator_list:
        if isinstance(decorator, ast.Name):
            yield decorator.id
        elif isinstance(decorator, ast.Attribute):
            yield decorator.attr
        elif isinstance(decorator, ast.Call):
            if isinstance(decorator.func, ast.Name):
                yield decorator.func.id
            else:
                yield decorator.func.attr
        else:
            assert False, decorator


def is_stub_function(function: FunctionTypes) -> bool:
    if isinstance(function, ast.Lambda):
        return isinstance(function.body, ast.Ellipsis)

    if (not isinstance(function, ast.Lambda)) and len(function.body) > 1:
        return False

    statement = function.body[0]
    if isinstance(statement, ast.Pass):
        return True
    if isinstance(statement, ast.Expr) and isinstance(statement.value, ast.Ellipsis):
        return True
    # FIXME: ignore if the function is raise NotImplementedError()

    return False


class FunctionFinder(NodeVisitor):
    functions: List[FunctionTypes]

    def __init__(self) -> None:
        super().__init__()
        self.functions = []

    def visit_AsyncFunctionDef(self, function: ast.AsyncFunctionDef) -> None:
        self.functions.append(function)

    def visit_FunctionDef(self, function: ast.FunctionDef) -> None:
        self.functions.append(function)

    def visit_Lambda(self, function: ast.Lambda) -> None:
        self.functions.append(function)



import ast
import textwrap
from contextlib import nullcontext
from unittest.mock import patch

import pytest


@pytest.mark.parametrize("function, expected_names", [
    ("def foo(a, b, c): pass", ["a", "b", "c"]),
    ("def foo(a, b, *, c): pass", ["a", "b", "c"]),
    ("def foo(a, b, *, c=5): pass", ["a", "b", "c"]),
    ("def foo(*args): pass", ["args"]),
    ("def foo(**kwargs): pass", ["kwargs"]),
    ("def foo(a, b, *args, c, d=5, e, **kwargs): pass", ["a", "b", "args", "c", "d", "e", "kwargs"]),
    ("async def foo(a, b, c): pass", ["a", "b", "c"]),
    ("async def foo(a, b, *, c): pass", ["a", "b", "c"]),
    ("async def foo(a, b, *, c=5): pass", ["a", "b", "c"]),
    ("async def foo(*args): pass", ["args"]),
    ("async def foo(**kwargs): pass", ["kwargs"]),
    ("async def foo(a, b, *args, c, d=5, e, **kwargs): pass", ["a", "b", "args", "c", "d", "e", "kwargs"]),
    ("""
        class foo:
            def bar(self, cool):
                pass
    """, ["self", "cool"]),
    ("l = lambda g: 5", ["g"]),
])
def test_get_argument_names(function, expected_names):
    from flake8_unused_arguments import get_argument_names

    argument_names = get_argument_names(get_function(function))
    print(argument_names)
    print(expected_names)
    assert argument_names == expected_names


@pytest.mark.parametrize("function, expected_names", [
    ("def foo(a, b, c): return a + b", ["c"]),
    ("""
        class foo:
            def bar(self, cool):
                self.thing = cool
    """, []),
    ("""
        def external(a, b, c):
            def internal():
                a + b
    """, ["c"]),
    ("l = lambda g: 5", ["g"]),
])
def test_get_unused_arguments(function, expected_names):
    from flake8_unused_arguments import get_unused_arguments

    argument_names = get_unused_arguments(get_function(function))
    print(argument_names)
    print(expected_names)
    assert argument_names == expected_names


@pytest.mark.parametrize("function, expected_result", [
    ("""
    @a
    @thing.b
    @thing.c()
    @d()
    def foo():
        pass
    """, ["a", "b", "c", "d"]),
    ("lambda g: 5", []),
])
def test_get_decorator_names(function, expected_result):
    from flake8_unused_arguments import get_decorator_names

    function_names = list(get_decorator_names(get_function(function)))
    assert function_names == expected_result


@pytest.mark.parametrize("function, expected_result", [
    ("def foo(): pass", True),
    ("def foo(): ...", True),
    ("def foo(): return 5", False),
    ("lambda: ...", True),
    ("lambda: 5", False),
    ("""
    def foo():
        a = 5
        return 5
    """, False),
    ("""
    def foo():
        if 5:
            return
        else:
            return
    """, False),
])
def test_is_stub_function(function, expected_result):
    from flake8_unused_arguments import is_stub_function
    assert is_stub_function(get_function(function)) == expected_result


@pytest.mark.parametrize("function, options, expected_warnings", [
    ("""
    @abstractmethod
    def foo(a):
        pass
    """, {"ignore_abstract": False}, [(2, 0, "U100 Unused argument 'a'", 'unused argument')]),
    ("""
    @abstractmethod
    def foo(a):
        pass
    """, {"ignore_abstract": True}, []),
    ("""
    def foo(a):
        pass
    """, {"ignore_stubs": False}, [(2, 0, "U100 Unused argument 'a'", 'unused argument')]),
    ("""
    def foo(a):
        pass
    """, {"ignore_stubs": True}, []),
    ("""
    def foo(*args):
        pass
    """, {"ignore_variadic_names": True}, []),
    ("""
    def foo(**kwargs):
        pass
    """, {"ignore_variadic_names": True}, []),
    ("""
    def foo(*args):
        pass
    """, {"ignore_variadic_names": False}, [(2, 0, "U100 Unused argument 'args'", 'unused argument')]),
    ("""
    def foo(**kwargs):
        pass
    """, {"ignore_variadic_names": False}, [(2, 0, "U100 Unused argument 'kwargs'", 'unused argument')]),
    ("""
    def foo(_a):
        pass
    """, {}, [(2, 0, "U101 Unused argument '_a'", 'unused argument')]),
    ("""
    def foo(self):
        pass
    """, {}, []),
    ("""
    @classmethod
    def foo(cls):
        pass
    """, {}, []),
])
def test_integration(function, options, expected_warnings):
    from flake8_unused_arguments import Plugin

    with patch.multiple(Plugin, **options) if options else nullcontext():
        plugin = Plugin(ast.parse(textwrap.dedent(function)))
        warnings = list(plugin.run())
        print(warnings)
        assert warnings == expected_warnings


def get_function(text):
    from flake8_unused_arguments import FunctionFinder
    finder = FunctionFinder()
    finder.visit(ast.parse(textwrap.dedent(text)))
    return finder.functions[0]
